#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <semaphore.h>
#include <time.h>

static int SECOND = 1000000;

/* Taille de la salle d'attente */
#define NB_PLACES 5

/* Lock pour le porte monnaie */
pthread_mutexattr_t mutex_attr;
pthread_mutex_t money_lock;
pthread_mutex_t loose_lock;

/* Les quatres semaphores */
sem_t places;
sem_t clients;
sem_t fauteille;

sem_t coupe;
sem_t client_pret;

/* Caisse du coiffeur */
int porte_monnaie=0;
int manque_a_gagner=0;
char * client_name;

void proc_coiffeur(void *ptr);
void proc_client(void *ptr);
int num_Digits(int num);

/**
 * Count the number of Digits in a unsigned long
 */
int num_Digits(int num){
	if ( num < 10 )
		return 1;
	else
		return 1+num_Digits(num/10);
}

void proc_coiffeur(void *ptr)
{
	int i,pm,mag;
	printf("#Coif: Ouverture du salon.\n");
	while(1){
		sem_getvalue(&clients, &i);
		if(i==0){// il n'y a pas de clients
			pthread_mutex_lock(&money_lock);
			pm=porte_monnaie;
			pthread_mutex_unlock(&money_lock);
			pthread_mutex_lock(&loose_lock);
			mag=manque_a_gagner;
			pthread_mutex_unlock(&loose_lock);
			printf("#Coif: je vais me reposer, %i euros... pour l'instant.\n",pm );
			printf("#Coif: manque_a_gagner %d euros... :(\n",mag );
			int pl,cl,f,co,cl_p;
			sem_getvalue(&places     , &pl);
			sem_getvalue(&clients    , &cl);
			sem_getvalue(&fauteille  , &f);
			sem_getvalue(&coupe      , &co);
			sem_getvalue(&client_pret, &cl_p);
			printf("pl=%d, cl=%d, f=%d, co=%d, cl_p=%d\n",pl,cl,f,co,cl_p);
		}
		sem_wait(&clients);
		sem_post(&fauteille);
		sem_wait(&client_pret);
		printf("\033[0;31m");
		printf("#Coif: je coiffe un client. %s\n", client_name);
		printf("\033[0m");
		usleep(0.5*SECOND);				
		printf("\033[0;34m");
		printf("#Coif: fini %s\n",client_name);
		printf("\033[0m");
		sem_post(&coupe);
	}
}

void proc_client(void *ptr)
{
	int i;
	char * name = (char*)ptr;
	printf("\033[1;32m");
	printf("@%s: Bonjour !\n",name);
	printf("\033[0m");
	sem_getvalue(&places, &i);
	printf("\t@%s: Il reste (%d/%d) places de libre.\n",name, i,NB_PLACES);
	if ( i > 0){
		sem_wait(&places);
		int pl,cl,f,co,cl_p;
		sem_getvalue(&places     , &pl);
		sem_getvalue(&clients    , &cl);
		sem_getvalue(&fauteille  , &f);
		sem_getvalue(&coupe      , &co);
		sem_getvalue(&client_pret, &cl_p);
		printf("\tpl=%d, cl=%d, f=%d, co=%d, cl_p=%d\n",pl,cl,f,co,cl_p);
		if (pl < NB_PLACES-1){
			printf("\t@%s: J'attends sur une chaise\n",name);
		}
		sem_post(&clients);
		sem_wait(&fauteille);
		sem_post(&places);
		client_name=name;
		sem_post(&client_pret);
		printf("\t@%s: Je m'assoie sur le fauteuil pour me faire coiffer.\n",name);
		pthread_mutex_lock(&money_lock);
		porte_monnaie+=10;
		pthread_mutex_unlock(&money_lock);
		sem_wait(&coupe);
		
	}else{
		pthread_mutex_lock(&loose_lock);
		manque_a_gagner+=10;
		pthread_mutex_unlock(&loose_lock);
		printf("\t@%s: je n'ai pas de place pour attendre ! je pars\n",name);
	}
}

int main(void)
{
	/* Thread du coiffeur */
	pthread_t coiffeur;
	/* Ptr sur thread pour les clients */
	pthread_t *ptr_sur_thread;
	int nb_client=0;
	char* c_name;
	/* Init des semaphores */
	sem_init(&places, 0, NB_PLACES);
	sem_init(&clients, 0, 0);
	sem_init(&fauteille, 0, 0);

	sem_init(&client_pret,0,0);
	sem_init(&coupe,0,0);	
	/* Init du lock */
	pthread_mutex_init(&money_lock, &mutex_attr);
	pthread_mutex_init(&loose_lock, &mutex_attr);
	
	/* On demarre le coiffeur */
	pthread_create(&coiffeur, NULL, (void*)&proc_coiffeur, NULL);
	
	srand(time(NULL));
	/* Quand le user fait un retour de chariot, on lance un client */
	while(1)
	{	
		getchar();
		int a = 4;
		int b = 12;
		int c1=50,c2=150;
		int i;
		int arrivee;
		int nb_client_j =  a + (int) ((double) rand() * (b - a + 1.0 ) / (RAND_MAX+1.0));
		if (nb_client_j < 0){nb_client_j*=-1;}
		printf( "\033[1;33m" );
		printf("nb_client_j=%d\n",nb_client_j);
		printf("\033[0m");
		for ( i = 0;i< nb_client_j;i++){
			nb_client++;
			c_name=malloc((num_Digits(nb_client)+7)*sizeof(char)+1);
			sprintf(c_name,"client-%d",nb_client);

			ptr_sur_thread=(void*)malloc(sizeof(pthread_t));
			arrivee = c1 + (int)((double) rand() * (c2 - c1 + 1.0 ) / (RAND_MAX+1.0));
			if (arrivee < 0){arrivee*=-1;}
			usleep(arrivee*1000);
			printf("Nouveau Client: %s, %d\n",c_name,arrivee);
			pthread_create((void*)&ptr_sur_thread, NULL, (void*)&proc_client, c_name);
		}
	}
	/* Quand le coiffeur meurt, on exit */
	pthread_join(coiffeur, NULL);
	return 0;
}
